﻿//Definition of the OAS problem
//Authors: Lei He (l.he@tudelft.nl), Arthur Guijt (arthur.guijt12@gmail.com), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Lining Xing (xing2999@qq.com), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
//Date: October 20, 2019
//License: CC-BY-NC

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Diagnostics;

namespace MemeAlg
{
    class OASProblem:problem
    {
        //The sequence-dependent setup times are stored in a n*n table
        public double[,] SetupTime;

        /// <summary>
        /// Initialization of the problem
        /// </summary>
        public void ProblemIntialize()
        {
            ProblemName = "OAS";
            IsCostLimit = false;
            IsMultiVTW = false;
            IsDueTime = true;
            IsIdleTime = true;
            DestroyList = new int[5] { 0, 1, 2, 3, 6 };
            RepairList = new int[2] { 0, 1 };
        }

        /// <summary>
        /// Read instances from files (original datasets)
        /// </summary>
        /// <param name="OrderNum"></param>
        /// <param name="Tao"></param>
        /// <param name="R"></param>
        /// <param name="InstanceNumber"></param>
        public void LoadInstance(string OrderNum, string Tao, string R, string InstanceNumber)
        {
            ProblemIntialize();

            StreamReader OrderReader = new StreamReader(@"...\...\...\...\Dataset_OAS\" + OrderNum + "orders\\Tao" + Tao + "\\R" + R + "\\Dataslack_" + OrderNum + "orders_Tao" + Tao + "R" + R + "_" + InstanceNumber + ".txt");
            this.Instance.OrderNum = int.Parse(OrderNum);

            SetupTime = new double[this.Instance.OrderNum + 2, this.Instance.OrderNum + 2];

            string OrderInfo = OrderReader.ReadLine();
            string[] oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);

            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                Order temporder = new Order();
                temporder.OrderID = i;
                temporder.Release = double.Parse(oi[i]);
                OrderList.Add(temporder);
            }

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
                OrderList[i].ProcessTime = double.Parse(oi[i]);

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
                OrderList[i].DueTime = double.Parse(oi[i]);

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                OrderList[i].Deadline = double.Parse(oi[i]);
                OrderList[i].Deadline -= OrderList[i].ProcessTime;//Change the deadline to the latest start time
            }

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                OrderList[i].Revenue = double.Parse(oi[i]);
                OrderList[i].UnitRevenue = OrderList[i].Revenue / (OrderList[i].ProcessTime + 0.001);
            }
            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
                OrderList[i].Weight = double.Parse(oi[i]);

            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                OrderInfo = OrderReader.ReadLine();
                oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                for (int j = 0; j < this.Instance.OrderNum + 2; j++)
                {
                    SetupTime[i, j] = double.Parse(oi[j]);
                }
            }
            OrderReader.Close();

            OrderList[OrderList.Count - 1].Release = OrderList[OrderList.Count - 1].Deadline;
            OrderList[OrderList.Count - 1].DueTime = OrderList[OrderList.Count - 1].Deadline;
            OrderList[0].IsAccept = true;
            OrderList[0].IsDummy = true;
            OrderList[OrderList.Count - 1].IsAccept = true;
            OrderList[OrderList.Count - 1].IsDummy = true;


            double totalsetup = 0;
            for (int i = 0; i < OrderList.Count; i++)
            {
                totalsetup = 0;
                OrderList[i].MinSetup = 100;
                for (int j = 0; j < OrderList.Count; j++)
                {
                    totalsetup += SetupTime[j, i];
                    if (SetupTime[j, i] < OrderList[i].MinSetup && i != j && j != OrderList.Count - 1)
                        OrderList[i].MinSetup = SetupTime[j, i];
                }
                OrderList[i].TotalSetup = totalsetup;
            }

            double avglength=0;
            foreach (Order a in OrderList)
                avglength += (a.Deadline - a.Release);
            avglength /= (OrderList.Count-2);
            avglength /= (OrderList[OrderList.Count - 1].Deadline);

            Instance.AvgWindowLength = avglength;

            foreach (Order a in OrderList)
                TotalRev += a.Revenue;
        }

        /// <summary>
        /// Read instances from files (new datasets)
        /// </summary>
        /// <param name="InstanceNumber"></param>
        /// <param name="OrderNum"></param>
        public void LoadInstance(string InstanceNumber, string OrderNum)
        {
            ProblemIntialize();

            StreamReader OrderReader = new StreamReader(@"...\...\...\...\Dataset_GEN\"  + InstanceNumber + ".txt");
            this.Instance.OrderNum = int.Parse(OrderNum);

            SetupTime = new double[this.Instance.OrderNum + 2, this.Instance.OrderNum + 2];

            string OrderInfo = OrderReader.ReadLine();
            string[] oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);

            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                Order temporder = new Order();
                temporder.OrderID = i;
                temporder.Release = double.Parse(oi[i]);
                OrderList.Add(temporder);
            }

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
                OrderList[i].ProcessTime = double.Parse(oi[i]);

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
                OrderList[i].DueTime = double.Parse(oi[i]);

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                OrderList[i].Deadline = double.Parse(oi[i]);
                OrderList[i].Deadline -= OrderList[i].ProcessTime;
            }

            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                OrderList[i].Revenue = double.Parse(oi[i]);
                OrderList[i].UnitRevenue = OrderList[i].Revenue / (OrderList[i].ProcessTime + 0.001);
            }
            OrderInfo = OrderReader.ReadLine();
            oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
                OrderList[i].Weight = double.Parse(oi[i]);

            for (int i = 0; i < this.Instance.OrderNum + 2; i++)
            {
                OrderInfo = OrderReader.ReadLine();
                oi = OrderInfo.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                for (int j = 0; j < this.Instance.OrderNum + 2; j++)
                {
                    SetupTime[i, j] = double.Parse(oi[j]);
                }
            }
            OrderReader.Close();

            OrderList[OrderList.Count - 1].Release = OrderList[OrderList.Count - 1].Deadline;
            OrderList[OrderList.Count - 1].DueTime = OrderList[OrderList.Count - 1].Deadline;
            OrderList[0].IsAccept = true;
            OrderList[0].IsDummy = true;
            OrderList[OrderList.Count - 1].IsAccept = true;
            OrderList[OrderList.Count - 1].IsDummy = true;


            double totalsetup = 0;
            for (int i = 0; i < OrderList.Count; i++)
            {
                totalsetup = 0;
                OrderList[i].MinSetup = 100;
                for (int j = 0; j < OrderList.Count; j++)
                {
                    totalsetup += SetupTime[j, i];
                    if (SetupTime[j, i] < OrderList[i].MinSetup && i != j && j != OrderList.Count - 1)
                        OrderList[i].MinSetup = SetupTime[j, i];
                }
                OrderList[i].TotalSetup = totalsetup;
            }

            double avglength = 0;
            foreach (Order a in OrderList)
                avglength += (a.Deadline - a.Release);
            avglength /= (OrderList.Count - 2);
            avglength /= (OrderList[OrderList.Count - 1].Deadline);

            Instance.AvgWindowLength = avglength;

            foreach (Order a in OrderList)
                TotalRev += a.Revenue;
        }

        /// <summary>
        /// Calculate the setup time
        /// </summary>
        /// <param name="Pre"></param>
        /// <param name="Cur"></param>
        /// <param name="preend">the end of pre order</param>
        /// <param name="folsta">the start of the current order</param>
        /// <returns></returns>
        public override double GetSetupTime(SelectedOrder Pre, SelectedOrder Cur, double preend, double folsta)
        {
            return SetupTime[Pre.Order.OrderID, Cur.Order.OrderID];
        }

        /// <summary>
        /// Get the earliest start time of the current order according to its precursor
        /// </summary>
        /// <param name="Pre"></param>
        /// <param name="Cur"></param>
        /// <returns></returns>
        public override double GetEarliestTime(SelectedOrder Pre, SelectedOrder Cur)
        {
            double EarliestTime = 0;
            double ST = GetSetupTime(Pre, Cur, Pre.RealEndTime, Cur.RealStartTime);
            EarliestTime = Math.Max(Pre.RealEndTime + ST, Cur.Order.Release + ST);
            return EarliestTime;
        }

        /// <summary>
        /// the earliest start time of the current order according to a temporary end time of its precursor
        /// </summary>
        /// <param name="PreTempEnd"></param>
        /// <param name="Pre"></param>
        /// <param name="Cur"></param>
        /// <returns></returns>
        public override double GetTempEarliestTime(double PreTempEnd, SelectedOrder Pre, SelectedOrder Cur)
        {
            double EarliestTime = 0;
            double ST = GetSetupTime(Pre, Cur, PreTempEnd, Cur.RealStartTime);
            EarliestTime = Math.Max(PreTempEnd + ST, Cur.Order.Release + ST);
            return EarliestTime;
        }

        /// <summary>
        /// Get the latest start time of the current order according to its successor
        /// </summary>
        /// <param name="SucStart"></param>
        /// <param name="Cur"></param>
        /// <param name="Suc"></param>
        /// <returns></returns>
        public override double GetLatestTime(double SucStart, SelectedOrder Cur, SelectedOrder Suc)
        {
            double LatestTime = 0;
            LatestTime = Math.Min(Cur.Order.Deadline, SucStart - SetupTime[Cur.Order.OrderID,Suc.Order.OrderID]-Cur.Order.ProcessTime);
            return LatestTime;
        }

        /// <summary>
        /// Get the latest start time of the current order that the order will receive no penalty
        /// </summary>
        /// <param name="SucStart"></param>
        /// <param name="Cur"></param>
        /// <param name="Suc"></param>
        /// <returns></returns>
        public override double GetDueLatestTime(double SucStart, SelectedOrder Cur, SelectedOrder Suc)
        {
            double LatestTime = 0;
            LatestTime = Math.Min(Cur.Order.DueTime - Cur.Order.ProcessTime, SucStart - SetupTime[Cur.Order.OrderID, Suc.Order.OrderID] - Cur.Order.ProcessTime);
            return LatestTime;
        }
    }
}
