﻿//The main program to run Sparrow
//Authors: Lei He (l.he@tudelft.nl), Arthur Guijt (arthur.guijt12@gmail.com), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Lining Xing (xing2999@qq.com), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
//Date: October 20, 2019
//License: CC-BY-NC

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace MemeAlg
{
    public partial class SparrowAlgorithm : Form
    {
        public SparrowAlgorithm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Run Sparrow on the original instance set. All the instances are run 10 times and the average value is calculated and stored.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OriginalIns_Click(object sender, EventArgs e)
        {
            List<double> BKUB = new List<double>();

            StreamReader or = new StreamReader(@"...\...\...\...\Dataset_OAS\OAS_TS_Solutions_UBs.txt");
            or.ReadLine();
            int count = 0;
            string instanceinfo = " ";
            string profitinfo;
            while (!or.EndOfStream)
            {
                if (count == 0)
                    instanceinfo = or.ReadLine();
                profitinfo = or.ReadLine();
                string[] pi = profitinfo.Split(new char[] { '\t' }, StringSplitOptions.RemoveEmptyEntries);
                BKUB.Add(double.Parse(pi[1]));
                count++;
                if (count == 10)
                    count = 0;
            }
            or.Close();

            string nowtime = System.DateTime.Now.ToString("yy-MM-dd HH-mm-ss");

            int runs = 10;
            int instance = 0;
            int[] ordernum = new int[6] { 10, 15, 20, 25, 50, 100 };
            int[] tau = new int[5] { 1, 3, 5, 7, 9 };
            int[] r = new int[5] { 1, 3, 5, 7, 9 };
            for (int orderno = 3; orderno < 6; orderno++)
                for (int tauno = 0; tauno < 5; tauno++)
                    for (int rno = 0; rno < 5; rno++)
                    {
                        int temp = orderno * 5 * 5 * 10 + tauno * 5 * 10 + rno * 10;

                        int maxite = 0;

                        List<Solution> GapList = new List<Solution>();
                        List<Solution> TimeList = new List<Solution>();

                        for (int i = 1; i <= 10; i++)
                        {
                            for (int j = 0; j < runs; j++)
                            {
                                instance = orderno * 5 * 5 * 10 + tauno * 5 * 10 + rno * 10 + i - 1;
                                MemeticAlgSolver newSolver = new MemeticAlgSolver();
                                OASProblem ProblemToSolve = new OASProblem();
                                maxite = newSolver.MaxIte;
                                ProblemToSolve.LoadInstance(ordernum[orderno].ToString(), tau[tauno].ToString(), r[rno].ToString(), i.ToString());

                                newSolver.Problem = ProblemToSolve;
                                List<double> time = new List<double>();
                                List<double> rev = new List<double>();
                                newSolver.StartExecute(out time, out rev);
                                List<double> revsl = new List<double>();
                                List<double> timesl = new List<double>();
                                for (int k = 0; k < 10; k++)
                                {
                                    double tempt, tempr;
                                    tempr = (BKUB[instance] * 1.0 - rev[(k + 1) * maxite / 10]) / BKUB[instance];
                                    tempt = time[(k + 1) * maxite / 10];
                                    revsl.Add(tempr);
                                    timesl.Add(tempt);
                                }
                                Solution revlist = new Solution();
                                revlist.sl = revsl;
                                Solution timelist = new Solution();
                                timelist.sl = timesl;

                                GapList.Add(revlist);
                                TimeList.Add(timelist);
                            }
                        }

                        StreamWriter solwriter = new StreamWriter(@"...\...\...\...\Dataset_OAS\" + ordernum[orderno].ToString() + "orders" + tau[tauno].ToString() + "tao" + r[rno].ToString() + "r_10runs.csv");

                        for (int i = 0; i < runs * 10; i++)
                        {
                            if (i % runs == 0)
                                solwriter.WriteLine(ordernum[orderno].ToString() + "orders" + tau[tauno].ToString() + "tao" + r[rno].ToString() + "r_" + i / runs);
                            solwriter.WriteLine(TimeList[i].sl[0].ToString("0.000000") + ',' + GapList[i].sl[0].ToString("0.000000") + ',' +
                                TimeList[i].sl[1].ToString("0.000000") + ',' + GapList[i].sl[1].ToString("0.000000") + ',' +
                                TimeList[i].sl[2].ToString("0.000000") + ',' + GapList[i].sl[2].ToString("0.000000") + ',' +
                                TimeList[i].sl[3].ToString("0.000000") + ',' + GapList[i].sl[3].ToString("0.000000") + ',' +
                                TimeList[i].sl[4].ToString("0.000000") + ',' + GapList[i].sl[4].ToString("0.000000") + ',' +
                                TimeList[i].sl[5].ToString("0.000000") + ',' + GapList[i].sl[5].ToString("0.000000") + ',' +
                                TimeList[i].sl[6].ToString("0.000000") + ',' + GapList[i].sl[6].ToString("0.000000") + ',' +
                                TimeList[i].sl[7].ToString("0.000000") + ',' + GapList[i].sl[7].ToString("0.000000") + ',' +
                                TimeList[i].sl[8].ToString("0.000000") + ',' + GapList[i].sl[8].ToString("0.000000") + ',' +
                                TimeList[i].sl[9].ToString("0.000000") + ',' + GapList[i].sl[9].ToString("0.000000"));
                        }

                        for (int it = 9; it < 10; it++)
                        {
                            StreamWriter stat = new StreamWriter(@"...\...\...\...\Dataset_OAS\" + nowtime + "_test.csv", true);

                            double bkubbest = 1, tbest = 100000, gap = 0;
                            List<double> timebest = new List<double>();
                            List<double> gapbest = new List<double>();
                            List<double> gapavg = new List<double>();

                            for (int i = 0; i < runs * 10; i++)
                            {
                                if (TimeList[i].sl[it] < tbest)
                                    tbest = TimeList[i].sl[it];
                                if (GapList[i].sl[it] < bkubbest)
                                    bkubbest = GapList[i].sl[it];
                                gap += GapList[i].sl[it];
                                if (i % runs == runs - 1)
                                {
                                    gap /= runs;
                                    gapavg.Add(gap);
                                    gap = 0;
                                    timebest.Add(tbest);
                                    gapbest.Add(bkubbest);
                                    bkubbest = 1; tbest = 100000;
                                }
                            }

                            stat.WriteLine(it.ToString() + ',' + ordernum[orderno].ToString() + "orders" + tau[tauno].ToString() + "tao" + r[rno].ToString() + "r," + gapavg.Min().ToString("0.000000") + ',' + gapavg.Average().ToString("0.000000") + ',' + gapavg.Max().ToString("0.000000") + ',' +
                                        gapbest.Min().ToString("0.000000") + ',' + gapbest.Average().ToString("0.000000") + ',' + gapbest.Max().ToString("0.000000") + ',' + timebest.Min().ToString("0.000000") + ',' + timebest.Average().ToString("0.000000") + ',' + timebest.Max().ToString("0.000000"));

                            stat.Close();
                        }
                        solwriter.Close();
                    }
        }

        /// <summary>
        /// Run Sparrow on the three new instance sets. All the instances are run 5 times and the average value is calculated and stored.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void NewIns_Click(object sender, EventArgs e)
        {
            string instancepath = "Commerce";
            int runs = 5;
            for (int i = 0; i <= 5; i++)
            {
                string temp1 = @"\" + (i * 2).ToString() + @"corr\";

                List<double> ProfitList10Ins = new List<double>();
                List<double> TimeList10Ins = new List<double>();

                StreamWriter solwriter = new StreamWriter(@"...\...\...\...\Dataset_GEN\Commerce.csv", true);
                StreamWriter solwritersingle = new StreamWriter(@"...\...\...\...\Dataset_GEN\Data_" + (i * 2).ToString() + ".csv");

                for (int m = 1; m <= 10; m++)
                {
                    string temp = "Data_" + (i * 2).ToString() + "corr_" + m.ToString();
                    List<double> ProfitList = new List<double>();
                    List<double> TimeList = new List<double>();
                    solwritersingle.WriteLine(temp);

                    for (int j = 0; j < runs; j++)
                    {
                        MemeticAlgSolver newSolver = new MemeticAlgSolver();
                        OASProblem ProblemToSolve = new OASProblem();
                        ProblemToSolve.LoadInstance(instancepath + temp1 + temp, 100.ToString());

                        newSolver.Problem = ProblemToSolve;
                        List<double> time = new List<double>();
                        List<double> rev = new List<double>();
                        newSolver.StartExecute(out time, out rev);
                        ProfitList.Add(rev[rev.Count - 1]);
                        TimeList.Add(time[time.Count - 1]);
                        solwritersingle.WriteLine(rev[rev.Count - 1].ToString("0.00") + ',' + time[time.Count - 1].ToString("0.00"));
                    }


                    ProfitList10Ins.Add(ProfitList.Average());
                    TimeList10Ins.Add(TimeList.Average());
                }

                solwriter.WriteLine(i.ToString() + ',' + ProfitList10Ins.Min().ToString("0.00") + ',' + ProfitList10Ins.Average().ToString("0.00") + ',' + ProfitList10Ins.Max().ToString("0.00")
                    + ',' + TimeList10Ins.Min().ToString("0.00") + ',' + TimeList10Ins.Average().ToString("0.00") + ',' + TimeList10Ins.Max().ToString("0.00"));
                solwriter.Close();
                solwritersingle.Close();
            }



            instancepath = "Satellite";
            runs = 5;
            for (int i = 2; i <= 6; i++)
            {
                string temp1 = @"\" + (i * 50).ToString() + @"orders\";

                List<double> ProfitList10Ins = new List<double>();
                List<double> TimeList10Ins = new List<double>();

                StreamWriter solwriter = new StreamWriter(@"...\...\...\...\Dataset_GEN\Satellite.csv", true);
                StreamWriter solwritersingle = new StreamWriter(@"...\...\...\...\Dataset_GEN\Satellite_Data_" + (i * 50).ToString() + ".csv");

                for (int m = 1; m <= 10; m++)
                {
                    string temp = "Data_" + (i * 50).ToString() + "orders_" + m.ToString();
                    List<double> ProfitList = new List<double>();
                    List<double> TimeList = new List<double>();
                    solwritersingle.WriteLine(temp);

                    for (int j = 0; j < runs; j++)
                    {
                        MemeticAlgSolver newSolver = new MemeticAlgSolver();
                        OASProblem ProblemToSolve = new OASProblem();
                        ProblemToSolve.LoadInstance(instancepath + temp1 + temp, (i * 50).ToString());

                        newSolver.Problem = ProblemToSolve;
                        List<double> time = new List<double>();
                        List<double> rev = new List<double>();
                        newSolver.StartExecute(out time, out rev);
                        ProfitList.Add(rev[rev.Count - 1]);
                        TimeList.Add(time[time.Count - 1]);
                        solwritersingle.WriteLine(rev[rev.Count - 1].ToString("0.00") + ',' + time[time.Count - 1].ToString("0.00"));
                    }


                    ProfitList10Ins.Add(ProfitList.Average());
                    TimeList10Ins.Add(TimeList.Average());
                }

                solwriter.WriteLine(i.ToString() + ',' + ProfitList10Ins.Min().ToString("0.00") + ',' + ProfitList10Ins.Average().ToString("0.00") + ',' + ProfitList10Ins.Max().ToString("0.00")
                    + ',' + TimeList10Ins.Min().ToString("0.00") + ',' + TimeList10Ins.Average().ToString("0.00") + ',' + TimeList10Ins.Max().ToString("0.00"));
                solwriter.Close();
                solwritersingle.Close();
            }

            instancepath = "Travelrepair";
            runs = 5;
            for (int i = 10; i <= 20; i++)
            {
                string temp1 = @"\" + (i / 10.0).ToString("0.0") + @"tmul\";

                List<double> ProfitList10Ins = new List<double>();
                List<double> TimeList10Ins = new List<double>();

                StreamWriter solwriter = new StreamWriter(@"...\...\...\...\Dataset_GEN\Travelrepair.csv", true);
                StreamWriter solwritersingle = new StreamWriter(@"...\...\...\...\Dataset_GEN\Data_travelrepair" + (i / 10.0).ToString() + ".csv");

                for (int m = 1; m <= 10; m++)
                {
                    string temp = "Data_" + (i / 10.0).ToString("0.0") + "tmul_" + m.ToString();
                    List<double> ProfitList = new List<double>();
                    List<double> TimeList = new List<double>();
                    solwritersingle.WriteLine(temp);

                    for (int j = 0; j < runs; j++)
                    {
                        MemeticAlgSolver newSolver = new MemeticAlgSolver();
                        OASProblem ProblemToSolve = new OASProblem();
                        ProblemToSolve.LoadInstance(instancepath + temp1 + temp, (i * 10).ToString());

                        newSolver.Problem = ProblemToSolve;
                        List<double> time = new List<double>();
                        List<double> rev = new List<double>();
                        newSolver.StartExecute(out time, out rev);
                        ProfitList.Add(rev[rev.Count - 1]);
                        TimeList.Add(time[time.Count - 1]);
                        solwritersingle.WriteLine(rev[rev.Count - 1].ToString("0.00") + ',' + time[time.Count - 1].ToString("0.00"));
                    }


                    ProfitList10Ins.Add(ProfitList.Average());
                    TimeList10Ins.Add(TimeList.Average());
                }

                solwriter.WriteLine(i.ToString() + ',' + ProfitList10Ins.Min().ToString("0.00") + ',' + ProfitList10Ins.Average().ToString("0.00") + ',' + ProfitList10Ins.Max().ToString("0.00")
                    + ',' + TimeList10Ins.Min().ToString("0.00") + ',' + TimeList10Ins.Average().ToString("0.00") + ',' + TimeList10Ins.Max().ToString("0.00"));
                solwriter.Close();
                solwritersingle.Close();
                i++;
            }

        }

    }

    public class Solution
    {
        public List<double> sl = new List<double>();
    }
}
